import uuid
from typing import Any, Dict, List, Optional, Tuple, Union

import httpx
from chainlit.logger import logger

from .base import (
    AppUser,
    ChainlitGraphQLClient,
    ConversationDict,
    ConversationFilter,
    ElementDict,
    MessageDict,
    PageInfo,
    PaginatedResponse,
    Pagination,
    PersistedAppUser,
)


class ChainlitCloudClient(ChainlitGraphQLClient):
    chainlit_server: str

    def __init__(self, api_key: str, chainlit_server="https://cloud.chainlit.io"):
        # Remove trailing slash
        chainlit_server = chainlit_server.rstrip("/")
        super().__init__(api_key=api_key, chainlit_server=chainlit_server)
        self.chainlit_server = chainlit_server

    async def create_app_user(self, app_user: AppUser) -> Optional[PersistedAppUser]:
        mutation = """
            mutation ($username: String!, $role: Role!, $tags: [String!], $provider: String, $image: String) {
                createAppUser(username: $username, role: $role, tags: $tags, provider: $provider, image: $image) {
                    id,
                    username,
                    role,
                    tags,
                    provider,
                    image,
                    createdAt
                }
            }
            """
        variables = app_user.to_dict()
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not create app user.")
            return None

        return PersistedAppUser.from_dict(res["data"]["createAppUser"])

    async def update_app_user(self, app_user: AppUser) -> Optional[PersistedAppUser]:
        mutation = """
            mutation ($username: String!, $role: Role!, $tags: [String!], $provider: String, $image: String) {
                updateAppUser(username: $username, role: $role, tags: $tags, provider: $provider, image: $image) {
                    id,
                    username,
                    role,
                    tags,
                    provider,
                    image,
                    createdAt
                }
            }
            """
        variables = app_user.to_dict()
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not update app user.")
            return None

        return PersistedAppUser.from_dict(res["data"]["updateAppUser"])

    async def get_app_user(self, username: str) -> Optional[PersistedAppUser]:
        query = """
             query ($username: String!) {
                getAppUser(username: $username) {
                    id,
                    username,
                    role,
                    tags,
                    provider,
                    image,
                    createdAt
                }
            }
            """
        variables = {"username": username}
        res = await self.query(query, variables)

        if self.check_for_errors(res):
            logger.warning("Could not get app user.")
            return None

        return PersistedAppUser.from_dict(res["data"]["getAppUser"])

    async def delete_app_user(self, username: str) -> bool:
        mutation = """
                mutation ($username: String!) {
                    deleteAppUser(username: $username) {
                        id,
                    }
                }
                """
        variables = {"username": username}
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not delete app user.")
            return False

        return True

    async def create_conversation(
        self, app_user_id: Optional[str], tags: Optional[List[str]]
    ) -> Optional[str]:
        mutation = """
        mutation ($appUserId: String, $tags: [String!]) {
            createConversation (appUserId: $appUserId, tags: $tags) {
                id
            }
        }
        """
        variables = {}  # type: Dict[str, Any]
        if app_user_id is not None:
            variables["appUserId"] = app_user_id

        if tags:
            variables["tags"] = tags

        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not create conversation.")
            return None

        return res["data"]["createConversation"]["id"]

    async def delete_conversation(self, conversation_id: str) -> bool:
        mutation = """
        mutation ($id: ID!) {
            deleteConversation(id: $id) {
                id
            }
        }
        """
        variables = {"id": conversation_id}
        res = await self.mutation(mutation, variables)
        self.check_for_errors(res, raise_error=True)

        return True

    async def get_conversation_author(self, conversation_id: str) -> Optional[str]:
        query = """
        query ($id: ID!) {
            conversation(id: $id) {
               appUser {
                   username
               }
            }
        }
        """
        variables = {
            "id": conversation_id,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)
        data = res.get("data")
        conversation = data.get("conversation") if data else None
        return (
            conversation["appUser"].get("username") if conversation["appUser"] else None
        )

    async def get_conversation(self, conversation_id: str) -> ConversationDict:
        query = """
        query ($id: ID!) {
            conversation(id: $id) {
                id
                createdAt
                tags
                metadata
                appUser {
                    id
                    username
                }
                messages {
                    id
                    isError
                    parentId
                    indent
                    author
                    content
                    waitForAnswer
                    humanFeedback
                    humanFeedbackComment
                    disableHumanFeedback
                    language
                    prompt
                    authorIsUser
                    createdAt
                }
                elements {
                    id
                    conversationId
                    type
                    name
                    mime
                    url
                    display
                    language
                    size
                    forIds
                }
            }
        }
        """
        variables = {
            "id": conversation_id,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        return res["data"]["conversation"]

    async def update_conversation_metadata(self, conversation_id: str, metadata: Dict):
        mutation = """mutation ($conversationId: ID!, $metadata: Json!) {
                    updateConversationMetadata(conversationId: $conversationId, metadata: $metadata) {
                            id
                    }
                }"""
        variables = {
            "conversationId": conversation_id,
            "metadata": metadata,
        }

        res = await self.mutation(mutation, variables)
        self.check_for_errors(res, raise_error=True)

        return True

    async def get_conversations(
        self, pagination: Pagination, filter: ConversationFilter
    ):
        query = """query (
        $first: Int
        $cursor: String
        $withFeedback: Int
        $username: String
        $search: String
    ) {
        conversations(
        first: $first
        cursor: $cursor
        withFeedback: $withFeedback
        username: $username
        search: $search
        ) {
        pageInfo {
            endCursor
            hasNextPage
        }
        edges {
            cursor
            node {
            id
            createdAt
            tags
            appUser {
                username
            }
            messages {
                content
            }
            }
        }
        }
    }"""

        variables = {
            "first": pagination.first,
            "cursor": pagination.cursor,
            "withFeedback": filter.feedback,
            "username": filter.username,
            "search": filter.search,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        conversations = []

        for edge in res["data"]["conversations"]["edges"]:
            node = edge["node"]
            conversations.append(node)

        page_info = res["data"]["conversations"]["pageInfo"]

        return PaginatedResponse(
            pageInfo=PageInfo(
                hasNextPage=page_info["hasNextPage"],
                endCursor=page_info["endCursor"],
            ),
            data=conversations,
        )

    async def set_human_feedback(
        self, message_id: str, feedback: int, feedbackComment: Optional[str]
    ) -> bool:
        mutation = """mutation ($messageId: ID!, $humanFeedback: Int!, $humanFeedbackComment: String) {
                    setHumanFeedback(messageId: $messageId, humanFeedback: $humanFeedback, humanFeedbackComment: $humanFeedbackComment) {
                            id
                            humanFeedback
                            humanFeedbackComment
                    }
                }"""
        variables = {
            "messageId": message_id,
            "humanFeedback": feedback,
        }
        if feedbackComment:
            variables["humanFeedbackComment"] = feedbackComment
        res = await self.mutation(mutation, variables)
        self.check_for_errors(res, raise_error=True)

        return True

    async def get_message(self):
        raise NotImplementedError

    async def create_message(self, variables: MessageDict) -> Optional[str]:
        mutation = """
        mutation ($id: ID!, $conversationId: ID!, $author: String!, $content: String!, $language: String, $prompt: Json, $isError: Boolean, $parentId: String, $indent: Int, $authorIsUser: Boolean, $disableHumanFeedback: Boolean, $waitForAnswer: Boolean, $createdAt: StringOrFloat) {
            createMessage(id: $id, conversationId: $conversationId, author: $author, content: $content, language: $language, prompt: $prompt, isError: $isError, parentId: $parentId, indent: $indent, authorIsUser: $authorIsUser, disableHumanFeedback: $disableHumanFeedback, waitForAnswer: $waitForAnswer, createdAt: $createdAt) {
                id
            }
        }
        """
        res = await self.mutation(mutation, variables)
        if self.check_for_errors(res):
            logger.warning("Could not create message.")
            return None

        return res["data"]["createMessage"]["id"]

    async def update_message(self, message_id: str, variables: MessageDict) -> bool:
        mutation = """
        mutation ($messageId: ID!, $author: String!, $content: String!, $parentId: String, $language: String, $prompt: Json, $disableHumanFeedback: Boolean) {
            updateMessage(messageId: $messageId, author: $author, content: $content, parentId: $parentId, language: $language,  prompt: $prompt, disableHumanFeedback: $disableHumanFeedback) {
                id
            }
        }
        """
        res = await self.mutation(mutation, dict(messageId=message_id, **variables))

        if self.check_for_errors(res):
            logger.warning("Could not update message.")
            return False

        return True

    async def delete_message(self, message_id: str) -> bool:
        mutation = """
        mutation ($messageId: ID!) {
            deleteMessage(messageId: $messageId) {
                id
            }
        }
        """
        res = await self.mutation(mutation, {"messageId": message_id})

        if self.check_for_errors(res):
            logger.warning("Could not delete message.")
            return False

        return True

    async def get_element(
        self, conversation_id: str, element_id: str
    ) -> Optional[ElementDict]:
        query = """query (
        $conversationId: ID!
        $id: ID!
    ) {
        element(
        conversationId: $conversationId,
        id: $id
        ) {
        id
        conversationId
        type
        name
        mime
        url
        display
        language
        size
        forIds
        }
    }"""

        variables = {
            "conversationId": conversation_id,
            "id": element_id,
        }
        res = await self.query(query, variables)
        self.check_for_errors(res, raise_error=True)

        return res["data"]["element"]

    async def create_element(self, variables: ElementDict) -> Optional[ElementDict]:
        mutation = """
        mutation ($conversationId: ID!, $type: String!, $name: String!, $display: String!, $forIds: [String!]!, $url: String, $objectKey: String, $size: String, $language: String, $mime: String) {
            createElement(conversationId: $conversationId, type: $type, url: $url, objectKey: $objectKey, name: $name, display: $display, size: $size, language: $language, forIds: $forIds, mime: $mime) {
                id,
                type,
                url,
                objectKey,
                name,
                display,
                size,
                language,
                forIds,
                mime
            }
        }
        """
        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not create element.")
            return None

        return res["data"]["createElement"]

    async def update_element(self, variables: ElementDict) -> Optional[ElementDict]:
        mutation = """
        mutation ($conversationId: ID!, $id: ID!, $forIds: [String!]!) {
            updateElement(conversationId: $conversationId, id: $id, forIds: $forIds) {
                id,
            }
        }
        """

        res = await self.mutation(mutation, variables)

        if self.check_for_errors(res):
            logger.warning("Could not update element.")
            return None

        return res["data"]["updateElement"]

    async def upload_element(
        self, content: Union[bytes, str], mime: str, conversation_id: Optional[str]
    ) -> Dict:
        id = str(uuid.uuid4())
        body = {"fileName": id, "contentType": mime}

        if conversation_id:
            body["conversationId"] = conversation_id

        path = "/api/upload/file"

        async with httpx.AsyncClient() as client:
            response = await client.post(
                f"{self.chainlit_server}{path}",
                json=body,
                headers=self.headers,
            )
            if response.status_code != 200:
                reason = response.text
                logger.error(f"Failed to sign upload url: {reason}")
                return {"object_key": None, "url": None}
            json_res = response.json()

        upload_details = json_res["post"]
        object_key = upload_details["fields"]["key"]
        signed_url = json_res["signedUrl"]

        # Prepare form data
        form_data = {}  # type: Dict[str, Tuple[Union[str, None], Any]]
        for field_name, field_value in upload_details["fields"].items():
            form_data[field_name] = (None, field_value)

        # Add file to the form_data
        # Note: The content_type parameter is not needed here, as the correct MIME type should be set in the 'Content-Type' field from upload_details
        form_data["file"] = (id, content)

        async with httpx.AsyncClient() as client:
            upload_response = await client.post(
                upload_details["url"],
                files=form_data,
            )
            try:
                upload_response.raise_for_status()
                url = f'{upload_details["url"]}/{object_key}'
                return {"object_key": object_key, "url": signed_url}
            except Exception as e:
                logger.error(f"Failed to upload file: {str(e)}")
                return {"object_key": None, "url": None}



from .anthropic import Anthropic
from .huggingface import HFFlanT5
from .openai import (
    AzureChatOpenAI,
    AzureOpenAI,
    ChatOpenAI,
    OpenAI,
)
from .vertexai import (
    ChatVertexAI,
    GenerationVertexAI,
)



import asyncio
import json
from typing import Any, Dict

from chainlit.action import Action
from chainlit.auth import get_current_user, require_login
from chainlit.config import config
from chainlit.context import init_ws_context
from chainlit.data import chainlit_client
from chainlit.logger import logger
from chainlit.message import ErrorMessage, Message
from chainlit.server import socket
from chainlit.session import WebsocketSession
from chainlit.telemetry import trace_event
from chainlit.types import UIMessagePayload
from chainlit.user_session import user_sessions


def restore_existing_session(sid, session_id, emit_fn, ask_user_fn):
    """Restore a session from the sessionId provided by the client."""
    if session := WebsocketSession.get_by_id(session_id):
        session.restore(new_socket_id=sid)
        session.emit = emit_fn
        session.ask_user = ask_user_fn
        trace_event("session_restored")
        return True
    return False


async def persist_user_session(conversation_id: str, metadata: Dict):
    if not chainlit_client:
        return

    await chainlit_client.update_conversation_metadata(
        conversation_id=conversation_id, metadata=metadata
    )


async def resume_conversation(session: WebsocketSession):
    if not chainlit_client or not session.user or not session.conversation_id:
        return

    conversation = await chainlit_client.get_conversation(
        conversation_id=session.conversation_id
    )

    author = (
        conversation["appUser"].get("username") if conversation["appUser"] else None
    )
    user_is_author = author == session.user.username

    if conversation and user_is_author:
        metadata = conversation["metadata"] or {}
        user_sessions[session.id] = metadata.copy()
        if chat_profile := metadata.get("chat_profile"):
            session.chat_profile = chat_profile
        if chat_settings := metadata.get("chat_settings"):
            session.chat_settings = chat_settings

        trace_event("conversation_resumed")

        return conversation


def load_user_env(user_env):
    # Check user env
    if config.project.user_env:
        # Check if requested user environment variables are provided
        if user_env:
            user_env = json.loads(user_env)
            for key in config.project.user_env:
                if key not in user_env:
                    trace_event("missing_user_env")
                    raise ConnectionRefusedError(
                        "Missing user environment variable: " + key
                    )
        else:
            raise ConnectionRefusedError("Missing user environment variables")
    return user_env


@socket.on("connect")
async def connect(sid, environ, auth):
    if not config.code.on_chat_start and not config.code.on_message:
        raise ConnectionRefusedError(
            "You need to configure at least an on_chat_start or an on_message callback"
        )

    user = None
    token = None
    try:
        # Check if the authentication is required
        if require_login():
            authorization_header = environ.get("HTTP_AUTHORIZATION")
            token = authorization_header.split(" ")[1] if authorization_header else None
            user = await get_current_user(token=token)
    except Exception as e:
        return False

    # Function to send a message to this particular session
    def emit_fn(event, data):
        if session := WebsocketSession.get(sid):
            if session.should_stop:
                session.should_stop = False
                raise InterruptedError("Task stopped by user")
        return socket.emit(event, data, to=sid)

    # Function to ask the user a question
    def ask_user_fn(data, timeout):
        if session := WebsocketSession.get(sid):
            if session.should_stop:
                session.should_stop = False
                raise InterruptedError("Task stopped by user")
        return socket.call("ask", data, timeout=timeout, to=sid)

    session_id = environ.get("HTTP_X_CHAINLIT_SESSION_ID")
    if restore_existing_session(sid, session_id, emit_fn, ask_user_fn):
        return True

    user_env_string = environ.get("HTTP_USER_ENV")
    user_env = load_user_env(user_env_string)
    WebsocketSession(
        id=session_id,
        socket_id=sid,
        emit=emit_fn,
        ask_user=ask_user_fn,
        user_env=user_env,
        user=user,
        token=token,
        chat_profile=environ.get("HTTP_X_CHAINLIT_CHAT_PROFILE"),
        conversation_id=environ.get("HTTP_X_CHAINLIT_CONVERSATION_ID"),
    )

    trace_event("connection_successful")
    return True


@socket.on("connection_successful")
async def connection_successful(sid):
    context = init_ws_context(sid)

    if context.session.restored:
        return

    if context.session.conversation_id and config.code.on_chat_resume:
        conversation = await resume_conversation(context.session)
        if conversation:
            context.session.has_user_message = True
            await config.code.on_chat_resume(conversation)
            await context.emitter.resume_conversation(conversation)
            return

    if config.code.on_chat_start:
        """Call the on_chat_start function provided by the developer."""
        await config.code.on_chat_start()


@socket.on("clear_session")
async def clean_session(sid):
    if session := WebsocketSession.get(sid):
        # Clean up the user session
        if session.id in user_sessions:
            user_sessions.pop(session.id)

        # Clean up the session
        session.delete()


@socket.on("disconnect")
async def disconnect(sid):
    session = WebsocketSession.get(sid)
    if config.code.on_chat_end and session:
        init_ws_context(session)
        """Call the on_chat_end function provided by the developer."""
        await config.code.on_chat_end()

    if session and session.conversation_id:
        await persist_user_session(session.conversation_id, session.to_persistable())

    async def disconnect_on_timeout(sid):
        await asyncio.sleep(config.project.session_timeout)
        if session := WebsocketSession.get(sid):
            # Clean up the user session
            if session.id in user_sessions:
                user_sessions.pop(session.id)
            # Clean up the session
            session.delete()

    asyncio.ensure_future(disconnect_on_timeout(sid))


@socket.on("stop")
async def stop(sid):
    if session := WebsocketSession.get(sid):
        trace_event("stop_task")

        init_ws_context(session)
        await Message(author="System", content="Task stopped by the user.").send()

        session.should_stop = True

        if config.code.on_stop:
            await config.code.on_stop()


async def process_message(session: WebsocketSession, payload: UIMessagePayload):
    """Process a message from the user."""
    try:
        context = init_ws_context(session)
        await context.emitter.task_start()
        message = await context.emitter.process_user_message(payload)

        if config.code.on_message:
            await config.code.on_message(message)
    except InterruptedError:
        pass
    except Exception as e:
        logger.exception(e)
        await ErrorMessage(
            author="Error", content=str(e) or e.__class__.__name__
        ).send()
    finally:
        await context.emitter.task_end()


@socket.on("ui_message")
async def message(sid, payload: UIMessagePayload):
    """Handle a message sent by the User."""
    session = WebsocketSession.require(sid)
    session.should_stop = False

    await process_message(session, payload)


async def process_action(action: Action):
    callback = config.code.action_callbacks.get(action.name)
    if callback:
        await callback(action)
    else:
        logger.warning("No callback found for action %s", action.name)


@socket.on("action_call")
async def call_action(sid, action):
    """Handle an action call from the UI."""
    init_ws_context(sid)

    action = Action(**action)

    await process_action(action)


@socket.on("chat_settings_change")
async def change_settings(sid, settings: Dict[str, Any]):
    """Handle change settings submit from the UI."""
    context = init_ws_context(sid)

    for key, value in settings.items():
        context.session.chat_settings[key] = value

    if config.code.on_settings_update:
        await config.code.on_settings_update(settings)



import chainlit as cl
from fastapi import HTTPException

from fastapi.responses import StreamingResponse

from chainlit.input_widget import Select, Slider, Tags
from chainlit.playground.provider import BaseProvider

vertexai_common_inputs = [
    Slider(
        id="temperature",
        label="Temperature",
        min=0.0,
        max=0.99,
        step=0.01,
        initial=0.2,
    ),
    Slider(
        id="max_output_tokens",
        label="Max Output Tokens",
        min=0.0,
        max=1024,
        step=1,
        initial=256,
    ),
]


class ChatVertexAIProvider(BaseProvider):
    async def create_completion(self, request):
        await super().create_completion(request)
        from vertexai.language_models import ChatModel, CodeChatModel

        self.validate_env(request=request)

        llm_settings = request.prompt.settings
        self.require_settings(llm_settings)

        prompt = self.create_prompt(request)
        model_name = llm_settings["model"]
        if model_name.startswith("chat-"):
            model = ChatModel.from_pretrained(model_name)
        elif model_name.startswith("codechat-"):
            model = CodeChatModel.from_pretrained(model_name)
        else:
            raise HTTPException(
                status_code=400,
                detail=f"This model{model_name} is not implemented.",
            )
        del llm_settings["model"]
        chat = model.start_chat()

        async def create_event_stream():
            for response in await cl.make_async(chat.send_message_streaming)(
                prompt[0].formatted, **llm_settings
            ):
                yield response.text

        return StreamingResponse(create_event_stream())


class GenerationVertexAIProvider(BaseProvider):
    async def create_completion(self, request):
        await super().create_completion(request)
        from vertexai.language_models import TextGenerationModel, CodeGenerationModel

        self.validate_env(request=request)

        llm_settings = request.prompt.settings
        self.require_settings(llm_settings)

        prompt = self.create_prompt(request)
        model_name = llm_settings["model"]
        if model_name.startswith("text-"):
            model = TextGenerationModel.from_pretrained(model_name)
        elif model_name.startswith("code-"):
            model = CodeGenerationModel.from_pretrained(model_name)
        else:
            raise HTTPException(
                status_code=400,
                detail=f"This model{model_name} is not implemented.",
            )
        del llm_settings["model"]

        async def create_event_stream():
            for response in await cl.make_async(model.predict_streaming)(
                prompt, **llm_settings
            ):
                yield response.text

        return StreamingResponse(create_event_stream())


gcp_env_vars = {"google_application_credentials": "GOOGLE_APPLICATION_CREDENTIALS"}

ChatVertexAI = ChatVertexAIProvider(
    id="chat-vertexai",
    env_vars=gcp_env_vars,
    name="ChatVertexAI",
    inputs=[
        Select(
            id="model",
            label="Model",
            values=["chat-bison", "codechat-bison"],
            initial_value="chat-bison",
        ),
        *vertexai_common_inputs,
    ],
    is_chat=True,
)

GenerationVertexAI = GenerationVertexAIProvider(
    id="generation-vertexai",
    env_vars=gcp_env_vars,
    name="GenerationVertexAI",
    inputs=[
        Select(
            id="model",
            label="Model",
            values=["text-bison", "code-bison"],
            initial_value="text-bison",
        ),
        *vertexai_common_inputs,
    ],
    is_chat=False,
)



import json
import uuid
from enum import Enum
from io import BytesIO
from typing import Any, ClassVar, Dict, List, Optional, TypeVar, Union, cast

import aiofiles
import filetype
from chainlit.client.base import ElementDict, ElementDisplay, ElementSize, ElementType
from chainlit.client.cloud import ChainlitCloudClient
from chainlit.context import context
from chainlit.data import chainlit_client
from chainlit.logger import logger
from chainlit.telemetry import trace_event
from pydantic.dataclasses import Field, dataclass
from syncer import asyncio

mime_types = {
    "text": "text/plain",
    "tasklist": "application/json",
    "plotly": "application/json",
}


@dataclass
class Element:
    # The type of the element. This will be used to determine how to display the element in the UI.
    type: ClassVar[ElementType]

    # The ID of the element. This is set automatically when the element is sent to the UI.
    id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    # Name of the element, this will be used to reference the element in the UI.
    name: Optional[str] = None
    # The URL of the element if already hosted somehwere else.
    url: Optional[str] = None
    # The S3 object key.
    object_key: Optional[str] = None
    # The local path of the element.
    path: Optional[str] = None
    # The byte content of the element.
    content: Optional[Union[bytes, str]] = None
    # Controls how the image element should be displayed in the UI. Choices are “side” (default), “inline”, or “page”.
    display: ElementDisplay = Field(default="side")
    # Controls element size
    size: Optional[ElementSize] = None
    # The ID of the message this element is associated with.
    for_ids: List[str] = Field(default_factory=list)
    # The language, if relevant
    language: Optional[str] = None
    # Mime type, infered based on content if not provided
    mime: Optional[str] = None

    def __post_init__(self) -> None:
        trace_event(f"init {self.__class__.__name__}")
        self.persisted = False

        if not self.url and not self.path and not self.content:
            raise ValueError("Must provide url, path or content to instantiate element")

    def to_dict(self) -> ElementDict:
        _dict = ElementDict(
            {
                "id": self.id,
                "type": self.type,
                "url": self.url or "",
                "name": self.name or "",
                "display": self.display,
                "objectKey": getattr(self, "object_key", None),
                "size": getattr(self, "size", None),
                "language": getattr(self, "language", None),
                "forIds": getattr(self, "for_ids", None),
                "mime": getattr(self, "mime", None),
                "conversationId": None,
            }
        )
        return _dict

    @classmethod
    def from_dict(self, _dict: Dict):
        if "image" in _dict.get("mime", ""):
            return Image(
                id=_dict.get("id", str(uuid.uuid4())),
                content=_dict.get("content"),
                name=_dict.get("name"),
                url=_dict.get("url"),
                display=_dict.get("display", "inline"),
                mime=_dict.get("mime"),
            )
        else:
            return File(
                id=_dict.get("id", str(uuid.uuid4())),
                content=_dict.get("content"),
                name=_dict.get("name"),
                url=_dict.get("url"),
                language=_dict.get("language"),
                display=_dict.get("display", "inline"),
                size=_dict.get("size"),
                mime=_dict.get("mime"),
            )

    async def with_conversation_id(self):
        _dict = self.to_dict()
        _dict["conversationId"] = await context.session.get_conversation_id()
        return _dict

    async def preprocess_content(self):
        pass

    async def load(self):
        if self.path:
            async with aiofiles.open(self.path, "rb") as f:
                self.content = await f.read()
        else:
            raise ValueError("Must provide path or content to load element")

    async def persist(self, client: ChainlitCloudClient) -> Optional[ElementDict]:
        if not self.url and self.content and not self.persisted:
            conversation_id = await context.session.get_conversation_id()
            upload_res = await client.upload_element(
                content=self.content,
                mime=self.mime or "",
                conversation_id=conversation_id,
            )
            self.url = upload_res["url"]
            self.object_key = upload_res["object_key"]
        element_dict = await self.with_conversation_id()

        asyncio.create_task(self._persist(element_dict))

        return element_dict

    async def _persist(self, element: ElementDict):
        if not chainlit_client:
            return

        try:
            if self.persisted:
                await chainlit_client.update_element(element)
            else:
                await chainlit_client.create_element(element)
                self.persisted = True
        except Exception as e:
            logger.error(f"Failed to persist element: {str(e)}")

    async def before_emit(self, element: Dict) -> Dict:
        return element

    async def remove(self):
        trace_event(f"remove {self.__class__.__name__}")
        await context.emitter.emit("remove_element", {"id": self.id})

    async def send(self, for_id: Optional[str] = None):
        if not self.content and not self.url and self.path:
            await self.load()

        await self.preprocess_content()

        if not self.mime:
            # Only guess the mime type when the content is binary
            self.mime = (
                mime_types[self.type]
                if self.type in mime_types
                else filetype.guess_mime(self.content)
            )

        if for_id and for_id not in self.for_ids:
            self.for_ids.append(for_id)

        # We have a client, persist the element
        if chainlit_client:
            element_dict = await self.persist(chainlit_client)
            if element_dict:
                self.id = element_dict["id"]

        elif not self.url and not self.content:
            raise ValueError("Must provide url or content to send element")

        emit_dict = cast(Dict, self.to_dict())

        # Adding this out of to_dict since the dict will be persisted in the DB
        emit_dict["content"] = self.content

        # Element was already sent
        if len(self.for_ids) > 1:
            trace_event(f"update {self.__class__.__name__}")
            await context.emitter.emit(
                "update_element",
                {"id": self.id, "forIds": self.for_ids},
            )
        else:
            trace_event(f"send {self.__class__.__name__}")
            emit_dict = await self.before_emit(emit_dict)
            await context.emitter.emit("element", emit_dict)


ElementBased = TypeVar("ElementBased", bound=Element)


@dataclass
class Image(Element):
    type: ClassVar[ElementType] = "image"

    size: ElementSize = "medium"


@dataclass
class Avatar(Element):
    type: ClassVar[ElementType] = "avatar"

    async def send(self):
        element = None

        if not self.content and not self.url and self.path:
            await self.load()

        if not self.url and not self.content:
            raise ValueError("Must provide url or content to send element")

        element = self.to_dict()

        # Adding this out of to_dict since the dict will be persisted in the DB
        element["content"] = self.content

        if element:
            trace_event(f"send {self.__class__.__name__}")
            element = await self.before_emit(element)
            await context.emitter.emit("element", element)


@dataclass
class Text(Element):
    """Useful to send a text (not a message) to the UI."""

    type: ClassVar[ElementType] = "text"

    content: bytes = b""
    language: Optional[str] = None

    async def before_emit(self, text_element):
        if "content" in text_element and isinstance(text_element["content"], bytes):
            text_element["content"] = text_element["content"].decode("utf-8")
        return text_element


@dataclass
class Pdf(Element):
    """Useful to send a pdf to the UI."""

    type: ClassVar[ElementType] = "pdf"


@dataclass
class Pyplot(Element):
    """Useful to send a pyplot to the UI."""

    # We reuse the frontend image element to display the chart
    type: ClassVar[ElementType] = "image"

    size: ElementSize = "medium"
    # The type is set to Any because the figure is not serializable
    # and its actual type is checked in __post_init__.
    figure: Any = None

    def __post_init__(self) -> None:
        from matplotlib.figure import Figure

        if not isinstance(self.figure, Figure):
            raise TypeError("figure must be a matplotlib.figure.Figure")

        options = {
            "dpi": 200,
            "bbox_inches": "tight",
            "backend": "Agg",
            "format": "png",
        }
        image = BytesIO()
        self.figure.savefig(image, **options)
        self.content = image.getvalue()

        super().__post_init__()


class TaskStatus(Enum):
    READY = "ready"
    RUNNING = "running"
    FAILED = "failed"
    DONE = "done"


@dataclass
class Task:
    title: str
    status: TaskStatus = TaskStatus.READY
    forId: Optional[str] = None

    def __init__(
        self,
        title: str,
        status: TaskStatus = TaskStatus.READY,
        forId: Optional[str] = None,
    ):
        self.title = title
        self.status = status
        self.forId = forId


@dataclass
class TaskList(Element):
    type: ClassVar[ElementType] = "tasklist"
    tasks: List[Task] = Field(default_factory=list, exclude=True)
    status: str = "Ready"
    name: str = "tasklist"
    content: str = "dummy content to pass validation"

    async def add_task(self, task: Task):
        self.tasks.append(task)

    async def update(self):
        await self.send()

    async def preprocess_content(self):
        # serialize enum
        tasks = [
            {"title": task.title, "status": task.status.value, "forId": task.forId}
            for task in self.tasks
        ]

        # store stringified json in content so that it's correctly stored in the database
        self.content = json.dumps(
            {
                "status": self.status,
                "tasks": tasks,
            },
            indent=4,
            ensure_ascii=False,
        )


@dataclass
class Audio(Element):
    type: ClassVar[ElementType] = "audio"


@dataclass
class Video(Element):
    type: ClassVar[ElementType] = "video"

    size: ElementSize = "medium"


@dataclass
class File(Element):
    type: ClassVar[ElementType] = "file"


@dataclass
class Plotly(Element):
    """Useful to send a plotly to the UI."""

    type: ClassVar[ElementType] = "plotly"

    size: ElementSize = "medium"
    # The type is set to Any because the figure is not serializable
    # and its actual type is checked in __post_init__.
    figure: Any = None
    content: str = ""

    def __post_init__(self) -> None:
        from plotly import graph_objects as go
        from plotly import io as pio

        if not isinstance(self.figure, go.Figure):
            raise TypeError("figure must be a plotly.graph_objects.Figure")

        self.figure.layout.autosize = True
        self.figure.layout.width = None
        self.figure.layout.height = None
        self.content = pio.to_json(self.figure, validate=True)
        self.mime = "application/json"

        super().__post_init__()



from typing import Union

from fastapi.responses import StreamingResponse

from chainlit.playground.provider import BaseProvider
from chainlit.prompt import PromptMessage
from chainlit.sync import make_async

from chainlit import input_widget


class LangchainGenericProvider(BaseProvider):
    from langchain.chat_models.base import BaseChatModel
    from langchain.llms.base import LLM

    llm: Union[LLM, BaseChatModel]

    def __init__(
        self,
        id: str,
        name: str,
        llm: Union[LLM, BaseChatModel],
        is_chat: bool = False,
    ):
        super().__init__(
            id=id,
            name=name,
            env_vars={},
            inputs=[],
            is_chat=is_chat,
        )
        self.llm = llm

    def prompt_message_to_langchain_message(self, message: PromptMessage):
        from langchain.schema.messages import (
            AIMessage,
            FunctionMessage,
            HumanMessage,
            SystemMessage,
        )

        content = "" if message.formatted is None else message.formatted
        if message.role == "user":
            return HumanMessage(content=content)
        elif message.role == "assistant":
            return AIMessage(content=content)
        elif message.role == "system":
            return SystemMessage(content=content)
        elif message.role == "function":
            return FunctionMessage(
                content=content, name=message.name if message.name else "function"
            )
        else:
            raise ValueError(f"Got unknown type {message}")

    def format_message(self, message, prompt):
        message = super().format_message(message, prompt)
        return self.prompt_message_to_langchain_message(message)

    def message_to_string(self, message: PromptMessage) -> str:
        return message.to_string()

    async def create_completion(self, request):
        from langchain.schema.messages import BaseMessageChunk

        await super().create_completion(request)

        messages = self.create_prompt(request)

        stream = make_async(self.llm.stream)

        result = await stream(
            input=messages,
        )

        def create_event_stream():
            try:
                for chunk in result:
                    if isinstance(chunk, BaseMessageChunk):
                        yield chunk.content
                    else:
                        yield chunk
            except Exception as e:
                # The better solution would be to return a 500 error, but
                # langchain raises the error in the stream, and the http
                # headers have already been sent.
                yield f"Failed to create completion: {str(e)}"

        return StreamingResponse(create_event_stream())



from enum import Enum
from typing import Dict, List, Literal, Optional, TypedDict, Union

from chainlit.client.base import ConversationFilter, MessageDict, Pagination
from chainlit.element import File
from chainlit.prompt import Prompt
from dataclasses_json import DataClassJsonMixin
from pydantic import BaseModel
from pydantic.dataclasses import dataclass

InputWidgetType = Literal[
    "switch", "slider", "select", "textinput", "tags", "numberinput"
]


@dataclass
class FileSpec(DataClassJsonMixin):
    accept: Union[List[str], Dict[str, List[str]]]
    max_files: int
    max_size_mb: int


@dataclass
class ActionSpec(DataClassJsonMixin):
    keys: List[str]


@dataclass
class AskSpec(DataClassJsonMixin):
    """Specification for asking the user."""

    timeout: int
    type: Literal["text", "file", "action"]


@dataclass
class AskFileSpec(FileSpec, AskSpec, DataClassJsonMixin):
    """Specification for asking the user a file."""


@dataclass
class AskActionSpec(ActionSpec, AskSpec, DataClassJsonMixin):
    """Specification for asking the user an action"""


class AskResponse(TypedDict):
    content: str
    author: str


class UIMessagePayload(TypedDict):
    message: MessageDict
    files: Optional[List[Dict]]


@dataclass
class AskFileResponse:
    name: str
    path: str
    size: int
    type: str
    content: bytes


class AskActionResponse(TypedDict):
    name: str
    value: str
    label: str
    description: str
    forId: str
    id: str
    collapsed: bool


class CompletionRequest(BaseModel):
    prompt: Prompt
    userEnv: Dict[str, str]


class UpdateFeedbackRequest(BaseModel):
    messageId: str
    feedback: Literal[-1, 0, 1]
    feedbackComment: Optional[str] = None


class DeleteConversationRequest(BaseModel):
    conversationId: str


class GetConversationsRequest(BaseModel):
    pagination: Pagination
    filter: ConversationFilter


class Theme(str, Enum):
    light = "light"
    dark = "dark"


@dataclass
class ChatProfile(DataClassJsonMixin):
    """Specification for a chat profile that can be chosen by the user at the conversation start."""

    name: str
    markdown_description: str
    icon: Optional[str] = None



import asyncio
from typing import Any, Dict, List, Optional

from chainlit.context import context_var
from chainlit.element import Text
from chainlit.message import Message
from chainlit.prompt import Prompt, PromptMessage
from llama_index.callbacks.base import BaseCallbackHandler
from llama_index.callbacks.schema import CBEventType, EventPayload
from llama_index.llms.base import ChatMessage, ChatResponse, CompletionResponse

DEFAULT_IGNORE = [
    CBEventType.CHUNKING,
    CBEventType.SYNTHESIZE,
    CBEventType.EMBEDDING,
    CBEventType.NODE_PARSING,
    CBEventType.QUERY,
    CBEventType.TREE,
]


class LlamaIndexCallbackHandler(BaseCallbackHandler):
    """Base callback handler that can be used to track event starts and ends."""

    def __init__(
        self,
        event_starts_to_ignore: List[CBEventType] = DEFAULT_IGNORE,
        event_ends_to_ignore: List[CBEventType] = DEFAULT_IGNORE,
    ) -> None:
        """Initialize the base callback handler."""
        self.context = context_var.get()
        self.event_starts_to_ignore = tuple(event_starts_to_ignore)
        self.event_ends_to_ignore = tuple(event_ends_to_ignore)

    def _restore_context(self) -> None:
        """Restore Chainlit context in the current thread

        Chainlit context is local to the main thread, and LlamaIndex
        runs the callbacks in its own threads, so they don't have a
        Chainlit context by default.

        This method restores the context in which the callback handler
        has been created (it's always created in the main thread), so
        that we can actually send messages.
        """
        context_var.set(self.context)

    def _get_parent_id(self) -> Optional[str]:
        """Get the parent message id"""
        if root_message := self.context.session.root_message:
            return root_message.id
        return None

    def on_event_start(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        parent_id: str = "",
        **kwargs: Any,
    ) -> str:
        """Run when an event starts and return id of event."""
        self._restore_context()
        asyncio.run(
            Message(
                content="",
                author=event_type,
                parent_id=self._get_parent_id(),
            ).send()
        )

        return event_id

    def on_event_end(
        self,
        event_type: CBEventType,
        payload: Optional[Dict[str, Any]] = None,
        event_id: str = "",
        **kwargs: Any,
    ) -> None:
        """Run when an event ends."""
        if payload is None:
            return

        self._restore_context()

        if event_type == CBEventType.RETRIEVE:
            sources = payload.get(EventPayload.NODES)
            if sources:
                elements = [
                    Text(name=f"Source {idx}", content=source.node.get_text())
                    for idx, source in enumerate(sources)
                ]
                source_refs = "\, ".join(
                    [f"Source {idx}" for idx, _ in enumerate(sources)]
                )
                content = f"Retrieved the following sources: {source_refs}"

                asyncio.run(
                    Message(
                        content=content,
                        author=event_type,
                        elements=elements,
                        parent_id=self._get_parent_id(),
                    ).send()
                )

        if event_type == CBEventType.LLM:
            formatted_messages = payload.get(
                EventPayload.MESSAGES
            )  # type: Optional[List[ChatMessage]]
            formatted_prompt = payload.get(EventPayload.PROMPT)
            response = payload.get(EventPayload.RESPONSE)

            if formatted_messages:
                messages = [
                    PromptMessage(role=m.role.value, formatted=m.content)  # type: ignore[arg-type]
                    for m in formatted_messages
                ]
            else:
                messages = None

            if isinstance(response, ChatResponse):
                content = response.message.content or ""
            elif isinstance(response, CompletionResponse):
                content = response.text
            else:
                content = ""

            asyncio.run(
                Message(
                    content=content,
                    author=event_type,
                    parent_id=self._get_parent_id(),
                    prompt=Prompt(
                        formatted=formatted_prompt,
                        messages=messages,
                        completion=content,
                    ),
                ).send()
            )

    def _noop(self, *args, **kwargs):
        pass

    start_trace = _noop
    end_trace = _noop



import os
from datetime import datetime, timedelta
from typing import Any, Dict

import jwt
from chainlit.client.cloud import AppUser
from chainlit.config import config
from chainlit.data import chainlit_client
from chainlit.oauth_providers import get_configured_oauth_providers
from fastapi import Depends, HTTPException
from fastapi.security import OAuth2PasswordBearer

reuseable_oauth = OAuth2PasswordBearer(tokenUrl="/login", auto_error=False)


def get_jwt_secret():
    return os.environ.get("CHAINLIT_AUTH_SECRET")


def ensure_jwt_secret():
    if require_login() and get_jwt_secret() is None:
        raise ValueError(
            "You must provide a JWT secret in the environment to use password authentication. Run `chainlit create-secret` to generate one."
        )


def is_oauth_enabled():
    return config.code.oauth_callback and len(get_configured_oauth_providers()) > 0


def require_login():
    return (
        config.code.password_auth_callback is not None
        or config.code.header_auth_callback is not None
        or is_oauth_enabled()
    )


def get_configuration():
    return {
        "requireLogin": require_login(),
        "passwordAuth": config.code.password_auth_callback is not None,
        "headerAuth": config.code.header_auth_callback is not None,
        "oauthProviders": get_configured_oauth_providers()
        if is_oauth_enabled()
        else [],
    }


def create_jwt(data: AppUser) -> str:
    to_encode = data.to_dict()  # type: Dict[str, Any]
    to_encode.update(
        {
            "exp": datetime.utcnow() + timedelta(minutes=60 * 24 * 15),  # 15 days
        }
    )
    encoded_jwt = jwt.encode(to_encode, get_jwt_secret(), algorithm="HS256")
    return encoded_jwt


async def authenticate_user(token: str = Depends(reuseable_oauth)):
    try:
        dict = jwt.decode(
            token,
            get_jwt_secret(),
            algorithms=["HS256"],
            options={"verify_signature": True},
        )
        del dict["exp"]
        app_user = AppUser(**dict)
    except Exception as e:
        raise HTTPException(status_code=401, detail="Invalid authentication token")

    if chainlit_client:
        try:
            persisted_app_user = await chainlit_client.get_app_user(app_user.username)
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))
        if persisted_app_user == None:
            raise HTTPException(status_code=401, detail="User does not exist")

        return persisted_app_user
    else:
        return app_user


async def get_current_user(token: str = Depends(reuseable_oauth)):
    if not require_login():
        return None

    return await authenticate_user(token)



import json
from contextlib import contextmanager

from chainlit.input_widget import Select, Slider, Tags
from chainlit.playground.provider import BaseProvider
from fastapi import HTTPException
from fastapi.responses import StreamingResponse


def stringify_function_call(function_call):
    if isinstance(function_call, dict):
        _function_call = function_call.copy()
    else:
        _function_call = {
            "arguments": function_call.arguments,
            "name": function_call.name,
        }

    if "arguments" in _function_call and isinstance(_function_call["arguments"], str):
        _function_call["arguments"] = json.loads(_function_call["arguments"])
    return json.dumps(_function_call, indent=4, ensure_ascii=False)


openai_common_inputs = [
    Slider(
        id="temperature",
        label="Temperature",
        min=0.0,
        max=2.0,
        step=0.01,
        initial=0.9,
        tooltip="Higher values like 0.8 will make the output more random, while lower values like 0.2 will make it more focused and deterministic.",
    ),
    Slider(
        id="max_tokens",
        label="Max Tokens",
        min=0.0,
        max=8000,
        step=1,
        initial=256,
        tooltip="The maximum number of tokens to generate in the chat completion.",
    ),
    Slider(
        id="top_p",
        label="Top P",
        min=0.0,
        max=1.0,
        step=0.01,
        initial=1.0,
        tooltip="An alternative to sampling with temperature, called nucleus sampling, where the model considers the results of the tokens with top_p probability mass. So 0.1 means only the tokens comprising the top 10% probability mass are considered.",
    ),
    Slider(
        id="frequency_penalty",
        label="Frequency Penalty",
        min=-2.0,
        max=2.0,
        step=0.01,
        initial=0.0,
        tooltip="Number between -2.0 and 2.0. Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim.",
    ),
    Slider(
        id="presence_penalty",
        label="Presence Penalty",
        min=-2.0,
        max=2.0,
        step=0.01,
        initial=0.0,
        tooltip="Number between -2.0 and 2.0. Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics.",
    ),
    Tags(
        id="stop",
        label="Stop Sequences",
        initial=[],
        tooltip="Up to 4 sequences where the API will stop generating further tokens.",
    ),
]


@contextmanager
def handle_openai_error():
    import openai

    try:
        yield
    except openai.APITimeoutError as e:
        raise HTTPException(
            status_code=408,
            detail=f"OpenAI API request timed out: {e}",
        )
    except openai.APIError as e:
        raise HTTPException(
            status_code=500,
            detail=f"OpenAI API returned an API Error: {e}",
        )
    except openai.APIConnectionError as e:
        raise HTTPException(
            status_code=503,
            detail=f"OpenAI API request failed to connect: {e}",
        )
    except openai.AuthenticationError as e:
        raise HTTPException(
            status_code=403,
            detail=f"OpenAI API request was not authorized: {e}",
        )
    except openai.PermissionDeniedError as e:
        raise HTTPException(
            status_code=403,
            detail=f"OpenAI API request was not permitted: {e}",
        )
    except openai.RateLimitError as e:
        raise HTTPException(
            status_code=429,
            detail=f"OpenAI API request exceeded rate limit: {e}",
        )


class ChatOpenAIProvider(BaseProvider):
    def format_message(self, message, prompt):
        message = super().format_message(message, prompt)
        return message.to_openai()

    async def create_completion(self, request):
        await super().create_completion(request)
        from openai import AsyncClient

        env_settings = self.validate_env(request=request)

        client = AsyncClient(api_key=env_settings["api_key"])

        llm_settings = request.prompt.settings

        self.require_settings(llm_settings)

        messages = self.create_prompt(request)

        if "stop" in llm_settings:
            stop = llm_settings["stop"]

            # OpenAI doesn't support an empty stop array, clear it
            if isinstance(stop, list) and len(stop) == 0:
                stop = None

            llm_settings["stop"] = stop

        if request.prompt.functions:
            llm_settings["functions"] = request.prompt.functions
            llm_settings["stream"] = False
        else:
            llm_settings["stream"] = True

        with handle_openai_error():
            response = await client.chat.completions.create(
                messages=messages,
                **llm_settings,
            )

        if llm_settings["stream"]:

            async def create_event_stream():
                async for part in response:
                    if part.choices and part.choices[0].delta.content:
                        token = part.choices[0].delta.content
                        yield token
                    else:
                        continue

        else:

            async def create_event_stream():
                message = response.choices[0].message
                if function_call := message.function_call:
                    yield stringify_function_call(function_call)
                else:
                    yield message.content or ""

        return StreamingResponse(create_event_stream())


class OpenAIProvider(BaseProvider):
    def message_to_string(self, message):
        return message.to_string()

    async def create_completion(self, request):
        await super().create_completion(request)
        from openai import AsyncClient

        env_settings = self.validate_env(request=request)

        client = AsyncClient(api_key=env_settings["api_key"])

        llm_settings = request.prompt.settings

        self.require_settings(llm_settings)

        prompt = self.create_prompt(request)

        if "stop" in llm_settings:
            stop = llm_settings["stop"]

            # OpenAI doesn't support an empty stop array, clear it
            if isinstance(stop, list) and len(stop) == 0:
                stop = None

            llm_settings["stop"] = stop

        llm_settings["stream"] = True

        with handle_openai_error():
            response = await client.completions.create(
                prompt=prompt,
                **llm_settings,
            )

        async def create_event_stream():
            async for part in response:
                if part.choices and part.choices[0].text:
                    token = part.choices[0].text
                    yield token
                else:
                    continue

        return StreamingResponse(create_event_stream())


class AzureOpenAIProvider(BaseProvider):
    def message_to_string(self, message):
        return message.to_string()

    async def create_completion(self, request):
        await super().create_completion(request)
        from openai import AsyncAzureOpenAI

        env_settings = self.validate_env(request=request)

        client = AsyncAzureOpenAI(
            api_key=env_settings["api_key"],
            api_version=env_settings["api_version"],
            azure_endpoint=env_settings["azure_endpoint"],
            azure_ad_token=self.get_var(request, "AZURE_AD_TOKEN"),
            azure_ad_token_provider=self.get_var(request, "AZURE_AD_TOKEN_PROVIDER"),
            azure_deployment=self.get_var(request, "AZURE_DEPLOYMENT"),
        )
        llm_settings = request.prompt.settings

        self.require_settings(llm_settings)

        prompt = self.create_prompt(request)

        if "stop" in llm_settings:
            stop = llm_settings["stop"]

            # OpenAI doesn't support an empty stop array, clear it
            if isinstance(stop, list) and len(stop) == 0:
                stop = None

            llm_settings["stop"] = stop

        llm_settings["stream"] = True

        with handle_openai_error():
            response = await client.completions.create(
                prompt=prompt,
                **llm_settings,
            )

        async def create_event_stream():
            async for part in response:
                if part.choices and part.choices[0].text:
                    token = part.choices[0].text
                    yield token
                else:
                    continue

        return StreamingResponse(create_event_stream())


class AzureChatOpenAIProvider(BaseProvider):
    def format_message(self, message, prompt):
        message = super().format_message(message, prompt)
        return message.to_openai()

    async def create_completion(self, request):
        await super().create_completion(request)
        from openai import AsyncAzureOpenAI

        env_settings = self.validate_env(request=request)

        client = AsyncAzureOpenAI(
            api_key=env_settings["api_key"],
            api_version=env_settings["api_version"],
            azure_endpoint=env_settings["azure_endpoint"],
            azure_ad_token=self.get_var(request, "AZURE_AD_TOKEN"),
            azure_ad_token_provider=self.get_var(request, "AZURE_AD_TOKEN_PROVIDER"),
            azure_deployment=self.get_var(request, "AZURE_DEPLOYMENT"),
        )

        llm_settings = request.prompt.settings

        self.require_settings(llm_settings)

        messages = self.create_prompt(request)

        if "stop" in llm_settings:
            stop = llm_settings["stop"]

            # OpenAI doesn't support an empty stop array, clear it
            if isinstance(stop, list) and len(stop) == 0:
                stop = None

            llm_settings["stop"] = stop

        llm_settings["model"] = env_settings["deployment_name"]

        if request.prompt.functions:
            llm_settings["functions"] = request.prompt.functions
            llm_settings["stream"] = False
        else:
            llm_settings["stream"] = True

        with handle_openai_error():
            response = await client.chat.completions.create(
                messages=messages,
                **llm_settings,
            )

        if llm_settings["stream"]:

            async def create_event_stream():
                async for part in response:
                    if part.choices and part.choices[0].delta.content:
                        token = part.choices[0].delta.content
                        yield token
                    else:
                        continue

        else:

            async def create_event_stream():
                message = response.choices[0].message
                if function_call := message.function_call:
                    yield stringify_function_call(function_call)
                else:
                    yield message.content or ""

        return StreamingResponse(create_event_stream())


openai_env_vars = {"api_key": "OPENAI_API_KEY"}

azure_openai_env_vars = {
    "api_key": "AZURE_OPENAI_API_KEY",
    "api_version": "AZURE_OPENAI_API_VERSION",
    "azure_endpoint": "AZURE_OPENAI_ENDPOINT",
    "deployment_name": "AZURE_OPENAI_DEPLOYMENT_NAME",
}

ChatOpenAI = ChatOpenAIProvider(
    id="openai-chat",
    env_vars=openai_env_vars,
    name="ChatOpenAI",
    inputs=[
        Select(
            id="model",
            label="Model",
            values=["gpt-3.5-turbo", "gpt-3.5-turbo-16k", "gpt-4", "gpt-4-32k"],
            initial_value="gpt-3.5-turbo",
        ),
        *openai_common_inputs,
    ],
    is_chat=True,
)

OpenAI = OpenAIProvider(
    id="openai",
    name="OpenAI",
    env_vars=openai_env_vars,
    inputs=[
        Select(
            id="model",
            label="Model",
            values=["text-davinci-003", "text-davinci-002"],
            initial_value="text-davinci-003",
        ),
        *openai_common_inputs,
    ],
    is_chat=False,
)


AzureChatOpenAI = AzureChatOpenAIProvider(
    id="azure-openai-chat",
    env_vars=azure_openai_env_vars,
    name="AzureChatOpenAI",
    inputs=openai_common_inputs,
    is_chat=True,
)

AzureOpenAI = AzureOpenAIProvider(
    id="azure",
    name="AzureOpenAI",
    env_vars=azure_openai_env_vars,
    inputs=openai_common_inputs,
    is_chat=False,
)



from typing import Optional

from chainlit.input_widget import Slider
from chainlit.playground.provider import BaseProvider
from chainlit.sync import make_async
from fastapi import HTTPException
from fastapi.responses import StreamingResponse
from pydantic.dataclasses import dataclass


@dataclass
class BaseHuggingFaceProvider(BaseProvider):
    repo_id: Optional[str] = None
    task = "text2text-generation"

    async def create_completion(self, request):
        await super().create_completion(request)
        from huggingface_hub.inference_api import InferenceApi

        env_settings = self.validate_env(request=request)
        llm_settings = request.prompt.settings
        self.require_settings(llm_settings)

        client = InferenceApi(
            repo_id=self.repo_id,
            token=env_settings["api_token"],
            task=self.task,
        )

        prompt = self.create_prompt(request)

        response = await make_async(client)(inputs=prompt, params=llm_settings)

        if "error" in response:
            raise HTTPException(
                status_code=500,
                detail=f"Error raised by inference API: {response['error']}",
            )
        if client.task == "text2text-generation":

            def create_event_stream():
                yield response[0]["generated_text"]

            return StreamingResponse(create_event_stream())
        else:
            raise HTTPException(status_code=400, detail="Unsupported task")


flan_hf_env_vars = {"api_token": "HUGGINGFACE_API_TOKEN"}

HFFlanT5 = BaseHuggingFaceProvider(
    id="huggingface_hub",
    repo_id="declare-lab/flan-alpaca-large",
    name="Flan Alpaca Large",
    env_vars=flan_hf_env_vars,
    inputs=[
        Slider(
            id="temperature",
            label="Temperature",
            min=0.0,
            max=1.0,
            step=0.01,
            initial=0.9,
        ),
        Slider(
            id="max_length",
            label="Completion max length",
            min=1.0,
            max=5000,
            step=1.0,
            initial=256,
        ),
    ],
    is_chat=False,
)
